from enum import StrEnum
from pathlib import Path
from typing import Sequence, Iterable

from pydantic import BaseModel

from twitter.utils import hidden_value, load_lines, write_lines


def format_cookies(cookies):
    """
    # Фикс непонятно чего..
    """
    for cookie in cookies:
        if cookie.get('sameSite') == 'no_restriction':
            cookie['sameSite'] = 'None'


class GoogleAccountStatus(StrEnum):
    UNKNOWN = "UNKNOWN"
    BAD_COOKIES = "BAD_COOKIES"
    RECOVERY_EMAIL_REQUIRED = "RECOVERY_EMAIL_REQUIRED"
    RECOVERY_REQUIRED = "RECOVERY_REQUIRED"
    CAPTCHA_REQUIRED = "CAPTCHA_REQUIRED"
    PHONE_VERIFICATION_REQUIRED = "PHONE_VERIFICATION_REQUIRED"
    BANNED = "BANNED"
    GOOD = "GOOD"

    def __str__(self):
        return self.value


class GoogleAccount(BaseModel):
    email:          str
    password:       str
    recovery_email: str | None = None
    # TODO Валидировать cookies
    cookies:        list | None = None
    status: GoogleAccountStatus = GoogleAccountStatus.UNKNOWN

    @property
    def hidden_password(self) -> str | None:
        return hidden_value(self.password) if self.password else None

    def __repr__(self):
        return f"{self.__class__.__name__}(email={self.email})"

    def __str__(self):
        return self.email


def from_file(
        filepath: Path | str,
        *,
        separator: str = ":",
        fields: Sequence[str] = ("email", "password", "recovery_email"),
) -> list[GoogleAccount]:
    """
    :param filepath: Путь до файла с данными об аккаунтах.
    :param separator: Разделитель между данными в строке.
    :param fields: Кортеж, содержащий имена полей в порядке их появления в строке.
     Должно содержать как минимум два поля - email и password: `("email", )`
    :return: Список аккаунтов.
    """
    accounts = []
    for line in load_lines(filepath):
        data = dict(zip(fields, line.split(separator)))
        data.update({key: None for key in data if not data[key]})
        accounts.append(GoogleAccount(**data))
    return accounts


def to_file(
        filepath: Path | str,
        accounts: Iterable[GoogleAccount],
        *,
        separator: str = ";",
        fields: Sequence[str] = ("email", "password", "recovery_email"),
):
    lines = []
    for account in accounts:
        account_data = []
        for field_name in fields:
            field = getattr(account, field_name)
            field = field if field is not None else ""
            account_data.append(field)
        lines.append(separator.join(account_data))
    write_lines(filepath, lines)



from datetime import datetime, timedelta

from pydantic import BaseModel


class AuthToken(BaseModel):
    auth_token: str
    refresh_token: str | None = None
    expires_at: datetime | None = None

    @classmethod
    def from_googleapis(cls, auth_token: str, refresh_token: str, expires_in: int):
        expires_at = datetime.now() + timedelta(seconds=expires_in)
        return cls(auth_token=auth_token, refresh_token=refresh_token, expires_at=expires_at)

    @property
    def is_expired(self) -> bool:
        return datetime.now() >= self.expires_at

    @property
    def short_auth_token(self) -> str:
        start = self.auth_token[:3]
        end = self.auth_token[-3:]
        return f"{start}**{end}"

    def __str__(self):
        return self.short_auth_token


from twitter.base import BaseClient

from .models import AuthToken
from .errors import (
    HTTPException,
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    GoogleAPIsServerError,
)


class GoogleAPIsClient(BaseClient):
    def __init__(self, key: str, **session_kwargs):
        super().__init__(**session_kwargs)
        self.key = key

    async def request(self, method, url, **kwargs):
        params = kwargs["params"] = kwargs.get("params") or {}
        params["key"] = self.key
        response = await self._session.request(method, url, **kwargs)
        data = response.json()

        if response.status_code == 400:
            raise BadRequest(response, data)

        if response.status_code == 401:
            raise Unauthorized(response, data)

        if response.status_code == 403:
            raise Forbidden(response, data)

        if response.status_code == 404:
            raise NotFound(response, data)

        if response.status_code >= 500:
            raise GoogleAPIsServerError(response, data)

        if not 200 <= response.status_code < 300:
            raise HTTPException(response, data)

        return response, data

    async def request_auth_data(
            self,
            provider_id: str,
            continue_uri: str,
            custom_parameter: dict = None,
    ) -> tuple[str, str]:
        """
        :return: auth_uri, session_id
        """
        url = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/createAuthUri"
        payload = {
            "providerId": provider_id,
            "continueUri": continue_uri,
            "customParameter": custom_parameter if custom_parameter else {},
        }
        headers = {
            'authority': 'www.googleapis.com',
            'x-client-version': 'Chrome/Handler/2.20.2/FirebaseCore-web'
        }
        response, data = await self.request("POST", url, headers=headers, json=payload)
        return data["authUri"], data["sessionId"]

    async def sign_in(self, request_uri: str, session_id: str) -> tuple[dict, AuthToken]:
        url = "https://identitytoolkit.googleapis.com/v1/accounts:signInWithIdp"
        payload = {
            "requestUri": request_uri,
            "sessionId": session_id,
            "returnSecureToken": True,
            "returnIdpCredential": True,
        }
        headers = {
            'authority': 'identitytoolkit.googleapis.com',
            'x-client-version': 'Chrome/JsCore/10.7.1/FirebaseCore-web',
        }
        response, data = await self.request("POST", url, headers=headers, json=payload)
        auth_token = AuthToken.from_googleapis(data["idToken"], data["refreshToken"], int(data["expiresIn"]))
        return data, auth_token

    async def request_account_info(self, auth_token: str):
        url = "https://identitytoolkit.googleapis.com/v1/accounts:lookup"
        payload = {"idToken": auth_token}
        headers = {
            'authority': 'identitytoolkit.googleapis.com',
            'x-client-version': 'Chrome/JsCore/10.7.1/FirebaseCore-web',
        }
        response, data = await self.request("POST", url, headers=headers, json=payload)
        return data

    async def refresh_auth_token(self, refresh_token: str) -> tuple[dict, AuthToken]:
        url = "https://securetoken.googleapis.com/v1/token"
        payload = f"grant_type=refresh_token&refresh_token={refresh_token}"
        headers = {
            'authority': 'securetoken.googleapis.com',
            'content-type': 'application/x-www-form-urlencoded',
        }
        response, data = await self.request("POST", url, headers=headers, json=payload)
        auth_token = AuthToken.from_googleapis(data["access_token"], data["refresh_token"], data["expires_in"])
        return data, auth_token
